/*
 * (C) 2006 by Harald Welte <laforge@netfilter.org>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 * This file implements userspace helper support for nf_conntrack.
 *
 * It is implemented mainly by using the existing nfnetlink_queue and
 * nfnetlink_conntrack mechanism.  Only a minimal additional nfnetlink_helper
 * subsystem was created to register helpers (helper wrappers) inside the
 * kernel.
 *
 * We need two queues per helper, one for the helpfn() function, the other
 * for the expectfn().  Since the queue space is 16bit, this still allows
 * for 32k different helpers on one system. "This should be enough for
 * everyone".
 *
 * TODO:
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <dlfcn.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <arpa/inet.h>
#include "linux_list.h"

#include <libnfnetlink/libnfnetlink.h>
#include <libnetfilter_queue/libnetfilter_queue.h>
#include <libnetfilter_conntrack/libnetfilter_conntrack.h>
#include <libnetfilter_cthelper/libnetfilter_cthelper.h>

#define NFH_BUFSIZE	4096

struct nfcth_handle {
	struct nfnl_handle *nfnlh;		/* our own */
	struct nfnl_subsys_handle *nfnlssh;	/* our own */

	struct nfct_handle *ct_h;	/* nfnetlink_conntrack */
	struct nfq_handle *q_h;		/* nfnetlink_queue */

	struct list_head helpers;	/* list of helpers */
	struct list_head expqueue_list;	/* list of expfn's */
};

/* wrapper structure around nfct_expect */
struct nfcth_expect {
	struct list_head list;

	struct nfct_expect exp;
	
	void (*expectfn)(struct nfct_conntrack *ct,
			 struct nfcth_expect *this);
};

/* nfnetlink_helper lowlevel stuff */

/* called when we get a packet for helpfn */
static int helpfn_queue_cb(struct nfq_q_handle *qh, struct nfgenmsg *nfmsg,
			   struct nfq_data *nfad, void *data)
{
	struct nfcth_helper *helper = (struct nfcth_helper *) data;

	/* FIXME: we want to know the master conntrack */
	return helper->help(nfad);
}

/* called when we need to call an expfn */
static int expfn_queue_cb(struct nfq_q_handle *qh, struct nfgenmsg *nfmsg,
			  struct nfq_data *nfad, void *data)
{
	struct expqueue_entry *ee = (struct expqueue_entry *) data;

	/* FIXME: we want to know which expectation was confirmed */
	ee->expfn(nfad);

	return 0;
}

static int nfcth_helper_new(struct nfcth_handle *hh, 
			    struct nfcth_helper *helper)
{
	char buf[NFH_BUFSIZE];
	struct nfnlhdr *req = (void *)buf;
	u_int8_t l3num = helper->tuple.l3protonum;
	u_int32_t max_expected, timeout;
	u_int16_t queue_id;
	
	memset(&buf, 0, sizeof(buf));

	nfnl_fill_hdr(hh->nfnlssh, &req->nlh, 0, l3num, 0, NFHNL_MSG_HELPER_NEW,
		      NLM_F_REQUEST|NLM_F_CREATE|NLM_F_ACK);

	nfct_build_tuple(req, sizeof(buf), &helper->tuple, NFHA_TUPLE);
	nfct_build_tuple(req, sizeof(buf), &helper->mask, NFHA_MASK);

	max_expected = htonl(helper->max_expected);
	nfnl_addattr_l(&req->nlh, sizeof(buf), NFHA_MAX_EXPECTED,
		       &max_expected, sizeof(u_int32_t));

	timeout = htonl(helper->timeout);
	nfnl_addattr_l(&req->nlh, sizeof(buf), NFHA_TIMEOUT, &timeout,
		       sizeof(u_int32_t));

	queue_id = htons(helper->queue_id);
	nfnl_addattr_l(&req->nlh, sizeof(buf), NFHA_QUEUENUM, &queue_id,
		       sizeof(u_int16_t));

	nfnl_addattr_l(&req->nlh, sizeof(buf), NFHA_NAME, (char *)helper->name,
		       strlen(helper->name));

	return nfnl_talk(hh->nfnlh, &req->nlh, 0, 0, NULL, NULL, NULL);
}

static int nfcth_helper_del(struct nfcth_helper *helper)
{
	int err;
	struct nfnlhdr *req;
	char buf[NFH_BUFSIZE];
	req = (void *)buf;
	u_int8_t l3num = helper->tuple.l3protonum;
	
	memset(&buf, 0, sizeof(buf));

	nfnl_fill_hdr(helper->h->nfnlssh, &req->nlh, 0, l3num, 0,
		      NFHNL_MSG_HELPER_DELETE,
		      NLM_F_ROOT|NLM_F_MATCH|NLM_F_REQUEST|NLM_F_ACK);

	nfnl_addattr_l(&req->nlh, sizeof(buf), NFHA_NAME, (char *)helper->name,
		       strlen(helper->name));

	err = nfnl_send(helper->h->nfnlh, &req->nlh);
	if (err < 0)
		return err;

	return nfnl_talk(helper->h->nfnlh, &req->nlh, 0, 0, NULL, NULL, NULL);
}


/* library internal housekeeping functions */

struct expqueue_entry {
	struct list_head list;
	void *expectfn;
	u_int16_t queue;

	struct nfq_q_handle *qh;
};

static LIST_HEAD(expqueue_list);

static int expqueue_resolve(struct nfcth_handle *h, void *expectfn)
{
	struct expqueue_entry *ee;
	int ret;

	list_for_each_entry(ee, &h->expqueue_list, list) {
		if (ee->expectfn == expectfn)
			return ee->queue;
	}

	ee = malloc(sizeof(ee));
	if (!ee)
		return -ENOMEM;

	ee->expectfn = expectfn;

	/* bind to a new (random) queue id */
	ret = nfq_create_queue_rnd(h->q_h, &ee->queue, &expfn_queue_cb, ee);
	if (ret < 0) {
		free(ee);
		return ret;
	}

	list_add(&ee->list, &h->expqueue_list);

	return ee->queue;
}


/* public API */

struct nfcth_expect *
nfcth_expect_alloc(struct nfct_tuple *master, struct nfct_tuple *tuple,
		   struct nfct_tuple *mask, u_int32_t timeout,
		   u_int32_t id,
		   void (*expectfn)(struct nfct_conntrack *ct,
				    struct nfcth_expect *this))
{
	struct nfcth_expect *hexp;

	hexp = malloc(sizeof(*hexp));
	if (!hexp)
		return NULL;

	hexp->exp = nfct_expect_alloc(master, tuple, mask, timeout, id);
	if (!hexp->exp) {
		free(hexp);
		return NULL;
	}

	return hexp;
}

int nfcth_expect_related(struct nfcth_helper *helper, struct nfcth_expect *exp)
{
	struct nfct_handle *cth = helper->h->ct_h;
	int ret;
	int exp_queue_nr;

	/* do we already have a queue number for this expectfn? */
	exp_queue_nr = expqueue_resolve(helper->h, exp->expectfn);
	if (exp_queue_nr < 0)
		return exp_queue_nr;

	exp->exp.expectfn_queue_id = exp_queue_nr;

	ret = nfct_create_expectation(cth, exp->exp);
}

void nfcth_unexpect_related(struct nfcth_expect *exp)
{
	struct nfct_handle *cth = helper->h->ct_h;
	int ret;

	ret = nfct_delete_expectation(cth, tuple, id);
}


int nfcth_helper_register(struct nfcth_handle *h, struct nfcth_helper *helper)
{
	int ret;

	helper->h = h;

	/* create new (random) queue */
	helper->qh = nfq_create_queue_rnd(h->q_h, &helper->queue_id, 
					  &helpfn_queue_cb, helper);
	if (!helper->qh)
		return -1;

	//? nfq_set_mode(helper->qh, COPY_PACKET, 0xffff);

	ret = nfcth_helper_new(h, helper);
	if (ret < 0) {
		nfq_destroy_queue(helper->qh);
		return ret;
	}

	list_add(&helper->list, &h->helpers);

	return 0;
}

void nfcth_helper_unregister(struct nfcth_helper *helper)
{
	list_del(&helper->list);

	/* unregister helper kernel part */
	nfcth_helper_del(helper);

	/* destroy the corresponding queue */
	nfq_destroy_queue(helper->qh);
}

struct nfcth_handle *nfcth_open_nfnl(struct nfnl_handle *nfnlh)
{
	struct nfcth_handle *h;

	h = malloc(sizeof(*h));
	if (!h)
		return NULL;

	h->nfnlh = nfnlh;
	if (!h->nfnlh)
		goto out_free;

	h->nfnlssh = nfnl_subsys_open(h->nfnlh, NFNL_SUBSYS_CTHELPER, 
				      NFHNL_MSG_MAX, 0);
	if (!h->nfnlssh)
		goto out_nfnl_close;

	h->q_h = nfq_open_nfnl(h->nfnlh);
	if (!h->q_h)
		goto out_nfnlss_close;

	h->ct_h = nfct_open_nfnl(h->nfnlh, 0, 0);
	if (!h->ct_h)
		goto out_nfq;

	INIT_LIST_HEAD(&h->expqueue_list);
	INIT_LIST_HEAD(&h->helpers);

	return h;

out_nfct:
	nfct_close(h->ct_h);
out_nfq:
	nfq_close(h->q_h);
out_nfnlss_close:
	nfnl_subsys_close(h->nfnlssh);
out_nfnl_close:
	nfnl_close(h->nfnlh);
out_free:
	free(h);
	return NULL;
}

struct nfcth_handle *nfcth_open(void)
{
	struct nfnl_handle *nfnlh = nfnl_open();
	struct nfcth_handle *nfcth;

	if (!nfnlh)
		return NULL;

	nfcth = nfcth_open_nfnl(nfnlh);
	if (!nfcth)
		nfnl_close(nfnlh);

	return nfcth;
}

void nfcth_close(struct nfcth_handle *hh)
{
	struct expqueue_entry *ee, *ee2;
	struct nfcth_helper *h, *h2;

	nfct_close(hh->ct_h);
	nfq_close(hh->q_h);

	list_for_each_entry_safe(h, h2, &hh->helpers, list) {
		nfcth_helper_unregister(h);	
		free(h); /* FIXME */
	}
	list_for_each_entry_safe(ee, ee2, &hh->expqueue_list, list) {
		list_del(&ee->list);
		free(ee);
	}
	free(hh);
}

struct nfnl_handle *nfcth_nfnlh(struct nfcth_handle *h)
{
	return h->nfnlh;
}

int nfcth_fd(struct nfcth_handle *h)
{
	return nfnl_fd(h->nfnlh);
}

int nfcth_handle_packet(struct nfcth_handle *h, char *buf, int len)
{
	return nfnl_handle_packet(h->nfnlh, buf, len);
}
